#include <stdio.h>
#include <stdarg.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include "config.h"

#ifndef HAVE_GETOPT
#include "getopt.c"
#endif

#define BLOCKSIZE 1024

int iselfline(char *line);

int main(int argc, char *argv[])
{
    int c;
    int pcorrupt = 0;
    long int size = 0, slide = 0, corrupt = 0;
    char *line = NULL;
    char *outfile = NULL;
    int opt;
    extern int optind;
    extern char *optarg;
    FILE *fileout;
    FILE *filein;

    while ((opt = getopt(argc, argv, "o:hvcf")) != EOF) {
	switch (opt) {
	case 'o':
	    outfile = optarg;
	    break;

	case 'h':
	    printf("elfstrip [-o output-file] [-c] [files] ...\n\n-o output-file  File to write output to. Default is stdout.\n-c              Write corrupt lines to stderr.\n");
	    exit(0);

	case 'v':
	    printf("elfstrip - " PACKAGE " " VERSION "\n");
	    exit(0);

	case 'c':
	    pcorrupt = 1;
	    break;

	}
    }

    /* open up file for output */
    if (outfile) {
	if (!(fileout = fopen(outfile, "w"))) {
	    fprintf(stderr, "File open error\n");
	    exit(1);
	}
    }
    else
	fileout = stdout;

    /* open up and read each file for input */
    opt = optind;
    if (opt == argc)
	filein = stdin;
    else
	filein = NULL;

    /* loop through each file */
    while (filein || opt < argc) {

	/* open the files for reading */
	if (!filein) {
	    if (!strcmp(argv[opt], "-"))
		filein = stdin;
	    else if (!(filein = fopen(argv[opt], "r"))) {
		perror(argv[opt]);
		exit(1);
	    }
	}

	/* get each line of text until EOF */
	while ((c = fgetc(filein)) != EOF) {
	    /* enough space in buffer? no, make some more */
	    if (slide >= size) {
		size += BLOCKSIZE;
		if (!(line = (char *) realloc(line, size))) {
		    fprintf(stderr, "Memory allocation error\n");
		    exit(1);
		}
	    }

	    /* save character to buffer */
	    line[slide++] = (char) c;

	    /* reached an end of line? if so, deal with the line */
	    if (c == 10) {

		/* the line may be terminated with a CR, if so chop it off
		 * otherwise replace the LF with a NULL to indicate end of string */
		if ((slide > 0) && (line[slide - 1] == 13))
		    line[slide - 2] = 0;
		else
		    line[slide - 1] = 0;

		/* is the line a comment? if so, ignore it */
		switch (line[0]) {
		case 0:
		case '\t':
		case ' ':
		case ';':
		case '!':
		case '{':{
			break;
		    }
		default:{
			if (iselfline(line)) {
			    printf("%s\n",line);
			}
			else {
			    if (pcorrupt)
				fprintf(stderr, "%s\n", line);
			    corrupt++;
			}
		    }
		}

		free(line);
		line = NULL;
		slide = 0;
		size = 0;
	    }

	}
	if (line)
	    free(line);

	/* close file (but only if the file was a real file) and loop */
	if (filein != stdin)
	    fclose(filein);
	filein = NULL;
	opt++;
    }

    /* close out output file */
    if (fileout != stdout)
	fclose(fileout);

    /* report corrupt line count */
    if (corrupt)
	fprintf(stderr, "%ld corrupt/incomplete lines ignored.\n", corrupt);

    exit(0);
}

/* get field number in a line */
/* buffer returned is size + 127 bytes so that it may be overwritten by the calling routine */
char *getfield(char *line, int field)
{
    char *buf;
    int start = 0, end = 0, slide = 0;
    int ws = 0;
    int br = 0;

    /* if ws is 0, we are finding text. if ws is not 0, we are finding whitespace */
    /* if br is not 0, we are inside a set of brackets */
    while (field && line[slide] && (line[slide] != 13)) {

	/* deal with brackets */
	if (line[slide] == '"') {
	    if (br)
		br = 0;
	    else
		br = 1;
	}

	/* find whitespace */
	switch (line[slide]) {
	case 8:
	case ' ':{
		if (!br && ws) {
		    ws = 0;
		}
		break;
	    }
	default:{
		if (!ws) {
		    field--;
		    ws = 1;
		    start = slide;
		}
	    }
	}
	slide++;
    }

    /* we found the start, now find the end */
    while (!end && line[slide] && (line[slide] != 13)) {

	/* find whitespace */
	switch (line[slide]) {
	case '"':{
		if (br)
		    br = 0;
		else
		    br = 1;
		break;
	    }
	case 8:
	case ' ':{
		if (!br) {
		    end = slide;
		}
		break;
	    }
	}
	slide++;
    }

    /* if !end, we found the last field */
    if (!end)
	end = slide;

    /* if field not found, return "" */
    if (field)
	start = end = 0;

    /* create a buffer for the field and return it - buffer contains an extra 127 bytes */
    if (!(buf = (char *) malloc(end - start + 128)))
	return (NULL);
    strncpy(buf, line + start, end - start);
    buf[end - start] = 0;
    return (buf);
}

void plustospace(char *str)
{
    register int x;

    for (x = 0; str[x]; x++)
	if (str[x] == '+')
	    str[x] = ' ';
}



#define DEFAULT_FORMAT "%[^ ] %[^ ] %[^ ] [%[^]]] \"%[^\"]\" %[^ ] %[^ ] \"%[^\"]\" \"%[^\"]\""

int iselfline(char *line)
{
    char *vp[10]; /* 8 options + 1 NULL plus one spare */
    int formatparms;
    int x;

    /* allocate some space for the strings to be stored in */
    x = 9;
    while (x >= 0) {
	if (!(vp[x]=(char *)malloc(strlen(line)+1))) {
	    fprintf(stderr, "Memory allocation error\n");
	    exit(1);
	}
	vp[x][0] = 0;
	x--;
    }

    /* do the scan */
    formatparms = sscanf(line, DEFAULT_FORMAT, vp[1], vp[2], vp[3], vp[4], vp[5], vp[6], vp[7], vp[8], vp[9]);

    /* free the space for the strings to be stored in */
    x = 9;
    while (x >= 0) {
	free(vp[x]);
	x--;
    }

    if (formatparms >= 6)
	return(1);
    else
	return(0);

}
